@@ -123,6 +123,7 @@ class Storage(Asset):
         Args:
             name (str): Unique name of the asset (asset parameter)
             node (Node): Node, the storage is located in (asset parameter)
+                         Two nodes may be defined in case input and output are located in different nodes [node_input, node_output]
             timegrid (Timegrid): Timegrid for discretization (asset parameter)
             wacc (float): Weighted average cost of capital to discount cash flows in target (asset parameter)
             size (float): maximum volume of commodity in storage.
@@ -144,11 +145,11 @@ class Storage(Asset):
         self.size = size
         self.start_level = start_level
         self.end_level= end_level
-        if start_level > size: raise ValueError('Asset --'+self.name+'--: start level must be smaller than the storage size')
+        assert start_level <= size, 'Storage --'+self.name+'--: start level must be <=  storage size'
         self.cap_in = cap_in
         self.cap_out = cap_out
-        if cap_in <0 : raise ValueError('Asset --'+self.name+'--: cap_in must not be negative')
-        if cap_out <0 : raise ValueError('Asset --'+self.name+'--: cap_out must not be negative')        
+        assert cap_in  >=0, 'Storage --'+self.name+'--: cap_in must not be negative'
+        assert cap_out >=0, 'Storage --'+self.name+'--: cap_out must not be negative'
         self.eff_in = eff_in
         self.inflow = inflow
         self.cost_out = cost_out
@@ -158,7 +159,7 @@ class Storage(Asset):
         self.block_size = None
         if block_size is not None:
             self.block_size = block_size # defines the block size (as pandas frequency)
-             
+        assert len(self.nodes)<=2, 'for storage only one or two nodes valid'
 
     def setup_optim_problem(self, prices: dict, timegrid:Timegrid = None, costs_only:bool = False) -> OptimProblem:
         """ Set up optimization problem for asset
@@ -176,8 +177,7 @@ class Storage(Asset):
         if not timegrid is None:
             self.set_timegrid(timegrid)
         # check: timegrid set?                    
-        if not hasattr(self, 'timegrid'): 
-            raise ValueError('Set timegrid of asset before creating optim problem. Asset: '+ self.name)
+        assert hasattr(self, 'timegrid'), 'Set timegrid of asset before creating optim problem. Asset: '+ self.name
 
         dt =  self.timegrid.restricted.dt
         n = self.timegrid.restricted.T # moved to Timegrid
@@ -193,8 +193,8 @@ class Storage(Asset):
             if not (len(price) == self.timegrid.T): # price vector must have right size for discretization
                 raise ValueError('Length of price array must be equal to length of time grid. Asset: '+ self.name)
         # separation into in/out needed?  Only one or two dispatch variables per time step
-        sep_needed =  (self.eff_in != 1) or (self.cost_in !=0) or (self.cost_out !=0)
-
+        # new separation reason: separate nodes in and out
+        sep_needed =  (self.eff_in != 1) or (self.cost_in !=0) or (self.cost_out !=0) or (len(self.nodes)==2)
         # cost_store -- costs for keeping quantity in storage
         # effectively, for each time step t we have:   cost_store * sum_{i<t}(disp_i)
         # and after summing ofer time steps t we get   cost_store * sum_t(disp_t * N_t)  
@@ -249,7 +249,10 @@ class Storage(Asset):
             aa = []
             for myd in indBlocks:
                 my_bool = self.timegrid.restricted.timepoints<=myd
-                aa.append(np.argwhere(my_bool)[-1,-1])
+                if any(my_bool):
+                    aa.append(np.argwhere(my_bool)[-1,-1])
+                else:
+                    aa.append(0)                    
                 if all(my_bool): break # stop early
             aa = np.asarray(aa)
             if aa[-1]!=n:
@@ -274,9 +277,15 @@ class Storage(Asset):
         mapping = pd.DataFrame()
         if sep_needed:
             mapping['time_step'] = np.hstack((self.timegrid.restricted.I, self.timegrid.restricted.I))
+            if len(self.nodes)==1:
+                mapping['node']      = self.nodes[0].name
+            else: # separate nodes in / out.
+                mapping['node']  = np.nan
+                mapping.loc[0:n,'node']      = self.nodes[0].name
+                mapping.loc[n:2*n,'node']      = self.nodes[1].name
         else:
             mapping['time_step'] = self.timegrid.restricted.I
-        mapping['node']      = self.nodes[0].name
+            mapping['node']      = self.nodes[0].name
         mapping['asset']     = self.name
         mapping['type']      = 'd'
         return OptimProblem(c=c,l=l, u=u, A=A, b=b, cType=cType, mapping = mapping)
@@ -558,8 +567,16 @@ def define_restr(my_take, my_type, my_n, map, timegrid, node = None):
     my_A     = sp.lil_matrix((0, my_n)) 
     my_b     = np.empty(shape=(0))
     my_cType = ''
+
+    # need to alter mapping - create copy
+    map = map.copy()
+    map.reset_index(inplace = True)
+    # several rows per variable? if not add missing column
+    if 'disp_factor' not in map.columns:
+        map['disp_factor'] = 1.
+
     for (s,e,v) in zip(my_take['start'], my_take['end'], my_take['values']):
-        I = [] # row with all zeros, no connections to restriction
+        I = [] # collect indices of rows that match
         for i, t in enumerate(timegrid.restricted.timepoints):
             if (s <= t) and (e > t):
                 if node is None:
@@ -569,7 +586,11 @@ def define_restr(my_take, my_type, my_n, map, timegrid, node = None):
         if not len(I) == 0: # interval could be outside timegrid, then omit restriction
             my_cType  += my_type
             a      = sp.lil_matrix((1,my_n))
-            a[0,I] = 1.   
+            # effectively adding all dispatches with corresponding factor in the selected node
+            # sum over all rows per variable
+            for my_i in I: # iterate over rows
+                a[0, map.loc[my_i, 'index']] += map.loc[my_i, 'disp_factor']
+            # ... with several rows per variable not valid any more:  a[0, map.loc[I, 'index'].values] = map.loc[I, 'disp_factor'].values
             my_A   = sp.vstack((my_A, a))
             # adjust quantity in case the restr. interval does not fully lie in timegrid
             # length of complete interval scaled down to interval within grid
@@ -709,7 +730,7 @@ class ExtendedTransport(Transport):
             Extension of transport with more complex restrictions:
        
             - time dependent capacity restrictions
-            - MinTake & MaxTake for a list of periods. With efficiency, min/maxTake refer to the quantity delivered from/to node 2 
+            - MinTake & MaxTake for a list of periods. With efficiency, min/maxTake refer to the quantity delivered FROM node 1 
 
         Examples
             - with min_cap = max_cap and a detailed time series
@@ -767,7 +788,10 @@ class ExtendedTransport(Transport):
             # assert right sizes
             assert ( (len(max_take['start'])== (len(max_take['end']))) and (len(max_take['start'])== (len(max_take['values']))) )
             # restricting only output node. 
-            A1, b1, c1 = define_restr(max_take, 'U', n, op.mapping, timegrid, node = self.node_names[1])
+            ## lower bound, since we have a different sign (transporting FROM node 0)
+            my_take = max_take.copy() # need to alter
+            my_take['values'] = -np.asarray(my_take['values'])
+            A1, b1, c1 = define_restr(my_take, 'L', n, op.mapping, timegrid, node = self.node_names[0])
             A     = sp.vstack((A, A1))
             b     = np.hstack((b, b1))
             cType = cType+c1
@@ -775,7 +799,10 @@ class ExtendedTransport(Transport):
             # assert right sizes
             assert ( (len(min_take['start'])== (len(min_take['end']))) and (len(min_take['start'])== (len(min_take['values']))) )
             # restricting only output node
-            A1, b1, c1 = define_restr(min_take, 'L', n, op.mapping, timegrid, node = self.node_names[1])            
+            ## lower bound, since we have a different sign (transporting FROM node 0)
+            my_take = min_take.copy() # need to alter
+            my_take['values'] = -np.asarray(my_take['values'])
+            A1, b1, c1 = define_restr(my_take, 'U', n, op.mapping, timegrid, node = self.node_names[0])            
             A     = sp.vstack((A, A1))
             b     = np.hstack((b, b1))
             cType = cType+c1        
